"use strict";
var Result;
(function (Result) {
    /**
     * Failures due to external events or systems that compromise the validity of the test.
     */
    Result[Result["COMPROMISED"] = 1] = "COMPROMISED";
    /**
     * Failure, due to some other exception.
     */
    Result[Result["ERROR"] = 2] = "ERROR";
    /**
     * Failure, due to an assertion error
     * For a test case, this means one of the tests in the test case failed.
     */
    Result[Result["FAILURE"] = 4] = "FAILURE";
    /**
     * The test step was not executed because a previous step in this test case failed.
     * A whole test case can be skipped using tags or annotations to indicate that it is currently "work-in-progress"
     */
    Result[Result["SKIPPED"] = 8] = "SKIPPED";
    /**
     * The test or test case was deliberately ignored.
     * Tests can be ignored via the @Ignore annotation in JUnit, for example.
     * Ignored tests are not considered the same as pending tests: a pending test is one that
     * has been specified, but the corresponding code is yet to be implemented, whereas an
     * ignored test can be a temporarily-deactivated test (during refactoring, for example).
     */
    Result[Result["IGNORED"] = 16] = "IGNORED";
    /**
     * A pending test is one that has been specified but not yet implemented.
     * In a JUnit test case, you can use the (Thucydides) @Pending annotation to mark this.
     * A pending test case is a test case that has at least one pending test.
     */
    Result[Result["PENDING"] = 32] = "PENDING";
    /**
     * The test or test case ran as expected.
     */
    Result[Result["SUCCESS"] = 64] = "SUCCESS";
    Result[Result["Failed"] = 7] = "Failed";
    Result[Result["Finished"] = 71] = "Finished";
})(Result = exports.Result || (exports.Result = {}));
class Tag {
    constructor(type, values = []) {
        this.type = type;
        this.values = values;
    }
    static from(text) {
        let [, type, values] = Tag.Pattern.exec(text);
        return new Tag(type, values.split(',').filter(_ => _.length > 0).map(_ => _.trim()));
    }
    get value() {
        return this.values.join(', ');
    }
}
Tag.Pattern = /^@([\w-]+)[:\s]?(.*)/i;
exports.Tag = Tag;
class Scene {
    constructor(name, category, path, tags = [], id = `${category}:${name}`) {
        this.name = name;
        this.category = category;
        this.path = path;
        this.tags = tags;
        this.id = id;
    }
    toString() {
        return `${this.name} (category: ${this.category}, path: ${this.path}${when(this.id !== this.name, ', id: ' + this.id)})`;
    }
}
exports.Scene = Scene;
class Activity {
    constructor(name, id = name) {
        this.name = name;
        this.id = id;
    }
    equals(another) {
        return this.name === another.name && this.id === another.id;
    }
    toString() {
        return `${this.name}${when(this.id !== this.name, ' (id: ' + this.id + ')')}`;
    }
}
exports.Activity = Activity;
class Outcome {
    constructor(subject, result, error) {
        this.subject = subject;
        this.result = result;
        this.error = error;
    }
    toString() {
        return `${this.subject.toString()} (result: ${Result[this.result]}${when(!!this.error, ', error:' + this.error)})`;
    }
}
exports.Outcome = Outcome;
class Photo {
    constructor(path) {
        this.path = path;
    }
}
exports.Photo = Photo;
class PhotoReceipt {
    constructor(activity, photo) {
        this.activity = activity;
        this.photo = photo;
    }
    toString() {
        return this.activity.toString();
    }
}
exports.PhotoReceipt = PhotoReceipt;
function when(condition, value) {
    return condition ? value : '';
}
//# sourceMappingURL=model.js.map