"use strict";
const domain_1 = require("../domain");
const file_system_1 = require("../io/file_system");
const serenity_1 = require("../serenity");
const path = require("path");
const _ = require("lodash");
const stack_trace_1 = require("stack-trace");
const md5_1 = require("ts-md5/dist/md5");
function jsonReporter(pathToReports = serenity_1.Default_Path_To_Reports) {
    return new JsonReporter(new file_system_1.FileSystem(pathToReports));
}
exports.jsonReporter = jsonReporter;
class JsonReporter {
    constructor(fs) {
        this.fs = fs;
    }
    assignTo(stage) {
        this.stage = stage;
        this.stage.manager.registerInterestIn(JsonReporter.Events_of_Interest, this);
    }
    notifyOf(event) {
        switch (event.constructor.name) {
            case domain_1.SceneStarts.name: return this.sceneStarts(event);
            case domain_1.ActivityStarts.name: return this.activityStarts(event);
            case domain_1.ActivityFinished.name: return this.activityFinished(event);
            case domain_1.SceneFinished.name: return this.sceneFinished(event);
            case domain_1.PhotoAttempted.name: return this.photoAttempted(event);
            default: break; // ignore any other events
        }
    }
    sceneStarts({ value, timestamp }) {
        this.rehearsal = new Rehearsal(value, timestamp);
    }
    activityStarts({ value, timestamp }) {
        this.rehearsal.activityStarts(value, timestamp);
    }
    activityFinished({ value, timestamp }) {
        this.rehearsal.activityFinished(value, timestamp);
    }
    photoAttempted({ value, timestamp }) {
        this.rehearsal.photoTaken(value, timestamp);
    }
    sceneFinished({ value, timestamp }) {
        let filename = `${md5_1.Md5.hashStr(value.subject.id)}.json`;
        this.rehearsal.finished(value, timestamp);
        this.stage.manager.informOfWorkInProgress(this.rehearsal.report().then(report => this.fs.store(filename, JSON.stringify(report))));
    }
}
JsonReporter.Events_of_Interest = [domain_1.SceneStarts, domain_1.ActivityStarts, domain_1.ActivityFinished, domain_1.PhotoAttempted, domain_1.SceneFinished];
exports.JsonReporter = JsonReporter;
class Rehearsal {
    constructor(scene, timestamp) {
        let sceneReport = new SceneReport(scene, timestamp);
        this.current = sceneReport;
        this.previous = sceneReport;
    }
    activityStarts(step, timestamp) {
        let activityReport = new ActivityReport(step, timestamp);
        this.current.append(activityReport);
        this.current = activityReport;
    }
    activityFinished(outcome, timestamp) {
        this.current.completedWith(outcome, timestamp);
        this.previous = this.current;
        this.current = this.current.parent;
    }
    finished(outcome, timestamp) {
        this.current.completedWith(outcome, timestamp);
    }
    photoTaken(receipt, timestamp) {
        [this.previous, this.current]
            .filter(report => report instanceof ActivityReport)
            .map(report => report)
            .filter(report => report.concerns(receipt.activity))
            .map(report => report.attachPhoto(receipt.photo));
    }
    report() {
        return this.current.toJSON();
    }
}
class SerenityReport {
    constructor(startTimestamp) {
        this.children = [];
        this.startedAt = startTimestamp;
    }
    append(stepExecutionReport) {
        let report = stepExecutionReport;
        report.parent = this;
        this.children.push(report);
    }
    completedWith(outcome, finishedAt) {
        this.result = outcome.result;
        this.error = outcome.error;
        this.duration = finishedAt - this.startedAt;
    }
    errorIfPresent() {
        if (!this.error) {
            return undefined; // so that the field is not rendered (that's what Serenity JVM expects for now)
        }
        return {
            errorType: this.error.name,
            message: this.error.message,
            stackTrace: this.stackTraceOf(this.error),
        };
    }
    mapAll(items, mapper = x => x) {
        let onlyDefined = item => item !== undefined;
        return Promise.all(items).then(all => all.filter(onlyDefined).map(mapper));
    }
    ifNotEmpty(list) {
        return !!list.length ? list : undefined;
    }
    // todo: extract a parser
    stackTraceOf(error) {
        let serenityCode = /node_modules[\\/]serenity/, onlyIfFound = index => index !== -1 ? index : undefined, firstSerenityStackFrame = (stack) => onlyIfFound(stack.findIndex(frame => !!serenityCode.exec(frame.getFileName()))), parsed = stack_trace_1.parse(error);
        return parsed.slice(0, firstSerenityStackFrame(parsed)).map(frame => {
            return {
                declaringClass: frame.getTypeName() || frame.getFunctionName() || '',
                methodName: frame.getMethodName() || frame.getFunctionName() || '',
                fileName: frame.getFileName(),
                lineNumber: frame.getLineNumber(),
            };
        });
    }
}
class SceneReport extends SerenityReport {
    constructor(scene, startTimestamp) {
        super(startTimestamp);
        this.scene = scene;
    }
    toJSON() {
        return this.mapAll(this.children.map(r => r.toJSON())).then(serialisedChildren => {
            return {
                name: this.scene.name,
                title: this.scene.name,
                description: '',
                tags: this.tagsFor(this.scene),
                // driver                               // todo: missing
                testSource: 'cucumber',
                startTime: this.startedAt,
                manual: false,
                duration: this.duration,
                result: domain_1.Result[this.result],
                testSteps: serialisedChildren,
                issues: this.issuesCoveredBy(this.scene),
                userStory: {
                    id: this.dashify(this.scene.category),
                    storyName: this.scene.category,
                    path: path.relative(process.cwd(), this.scene.path),
                    type: 'feature',
                },
                testFailureCause: this.errorIfPresent(),
            };
        });
    }
    dashify(name) {
        let dashified = name
            .replace(/([a-z])([A-Z])/g, '$1-$2')
            .replace(/[ \t\W]/g, '-')
            .replace(/^-+|-+$/g, '');
        return dashified.toLowerCase();
    }
    isAnIssue(tag) {
        return ['issue', 'issues'].indexOf(tag.type) !== -1;
    }
    issuesCoveredBy(scene) {
        const onlyIssueTags = this.isAnIssue, toIssueIds = tag => tag.values;
        return _.chain(scene.tags).filter(onlyIssueTags).map(toIssueIds).flatten().uniq().value();
    }
    tagsFor(scene) {
        const isAnIssue = this.isAnIssue;
        function serialise(tag) {
            const noValue = (t) => { return { name: t.type, type: 'tag' }; }, withValue = (t) => { return { name: t.values.join(','), type: t.type }; };
            return tag.values.length === 0
                ? noValue(tag)
                : withValue(tag);
        }
        function breakDownIssues(tag) {
            return isAnIssue(tag)
                ? tag.values.map(issueId => new domain_1.Tag('issue', [issueId]))
                : tag;
        }
        function featureTag(featureName) {
            return {
                name: featureName,
                type: 'feature',
            };
        }
        return _.chain(scene.tags)
            .map(breakDownIssues)
            .flatten()
            .map(serialise)
            .uniqBy('name')
            .concat(featureTag(scene.category))
            .value();
    }
}
class ActivityReport extends SerenityReport {
    constructor(activity, startTimestamp) {
        super(startTimestamp);
        this.activity = activity;
        this.promisedPhotos = [];
    }
    concerns(activity) {
        return this.activity.equals(activity);
    }
    attachPhoto(promisedPhoto) {
        this.promisedPhotos.push(promisedPhoto);
    }
    completedWith(outcome, finishedAt) {
        super.completedWith(outcome, finishedAt);
    }
    toJSON() {
        return this.mapAll(this.promisedPhotos, this.serialise).then(serialisedPhotos => {
            return this.mapAll(this.children.map(r => r.toJSON())).then(serialisedChildren => {
                return {
                    description: this.activity.name,
                    startTime: this.startedAt,
                    duration: this.duration,
                    result: domain_1.Result[this.result],
                    children: serialisedChildren,
                    exception: this.errorIfPresent(),
                    screenshots: this.ifNotEmpty(serialisedPhotos),
                };
            });
        });
    }
    serialise(photo) {
        return { screenshot: path.basename(photo.path) };
    }
}
//# sourceMappingURL=json_reporter.js.map