"use strict";
const _ = require("lodash");
const events_1 = require("../domain/events");
class StageManager {
    constructor(journal) {
        this.journal = journal;
        this.listeners = [];
        this.wip = [];
    }
    notifyOf(event) {
        this.journal.record(event);
        if (this.listeners[event.constructor.name]) {
            this.listeners[event.constructor.name].notify(event);
        }
        // for the "catch-all" crew members
        if (this.listeners[events_1.DomainEvent.name]) {
            this.listeners[events_1.DomainEvent.name].notify(event);
        }
    }
    informOfWorkInProgress(promise) {
        this.wip.push(promise);
        return promise;
    }
    allDone() {
        return Promise.all(this.wip);
    }
    registerInterestIn(eventsOfInterest, crewMember) {
        eventsOfInterest.forEach(eventType => {
            if (!this.listeners[eventType.name]) {
                this.listeners[eventType.name] = new CrewMembersCommunicationChannel();
            }
            this.listeners[eventType.name].register(crewMember);
        });
    }
    readTheJournal() {
        return this.journal.read();
    }
    readNewJournalEntriesAs(readerId) {
        return this.journal.readAs(readerId);
    }
}
exports.StageManager = StageManager;
class Journal {
    constructor() {
        this.events = [];
        this.bookmarks = {};
    }
    /**
     * Records an event and stores it in memory
     *
     * @param event
     */
    record(event) {
        this.events.push(event);
    }
    /**
     * Returns a list of all the events ever recorded
     *
     * @return {Event[]} a list of events
     */
    read() {
        return _.sortBy(_.cloneDeep(this.events), event => event.timestamp);
    }
    /**
     * Returns a list of all the events recorded since you last checked
     *
     * @return {Event[]} a list of events
     */
    readAs(readerId) {
        let events = this.read(), bookmark = this.bookmarks[readerId] || 0;
        this.bookmarks[readerId] = events.length;
        return events.slice(bookmark);
    }
}
exports.Journal = Journal;
class CrewMembersCommunicationChannel {
    constructor() {
        this.listeners = [];
    }
    register(listener) {
        this.listeners.push(listener);
    }
    notify(event) {
        this.listeners.forEach(listener => listener.notifyOf(_.cloneDeep(event)));
    }
}
//# sourceMappingURL=stage_manager.js.map